@@ -13,18 +13,24 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+from midonet.neutron._i18n import _LE, _LW
 from midonet.neutron.client import base as c_base
 from midonet.neutron.common import config  # noqa
+from midonet.neutron.common import constants as m_const
+from midonet.neutron.db import l3_db_midonet
 from midonet.neutron import extensions
 
 from neutron.api import extensions as neutron_extensions
 from neutron.common import constants as n_const
+from neutron.common import exceptions as n_exc
 from neutron.db import common_db_mixin
 from neutron.db import extraroute_db
+from neutron.db import l3_db
 # Import l3_dvr_db to get the config options required for FWaaS
 from neutron.db import l3_dvr_db  # noqa
-from neutron.db import l3_gwmode_db
-from neutron import i18n
+from neutron.extensions import l3
+from neutron.extensions import multiprovidernet as mpnet
+from neutron.extensions import providernet as pnet
 from neutron.plugins.common import constants
 from oslo_config import cfg
 from oslo_log import helpers as log_helpers
@@ -32,13 +38,11 @@ from oslo_log import log as logging
 from oslo_utils import excutils
 
 LOG = logging.getLogger(__name__)
-_LE = i18n._LE
-_LW = i18n._LW
 
 
 class MidonetL3ServicePlugin(common_db_mixin.CommonDbMixin,
                              extraroute_db.ExtraRoute_db_mixin,
-                             l3_gwmode_db.L3_NAT_db_mixin):
+                             l3_db_midonet.MidonetL3DBMixin):
 
     """
     Implements L3 Router service plugin for Midonet.
@@ -48,6 +52,7 @@ class MidonetL3ServicePlugin(common_db_mixin.CommonDbMixin,
 
     def __init__(self):
         super(MidonetL3ServicePlugin, self).__init__()
+        l3_db.subscribe()
 
         # Instantiate MidoNet API client
         self.client = c_base.load_client(cfg.CONF.MIDONET)
@@ -63,11 +68,37 @@ class MidonetL3ServicePlugin(common_db_mixin.CommonDbMixin,
         """Returns string description of the plugin."""
         return ("Midonet L3 Router Service Plugin")
 
+    @staticmethod
+    def _segments(network):
+        if pnet.NETWORK_TYPE in network:
+            yield {
+                pnet.NETWORK_TYPE: network[pnet.NETWORK_TYPE],
+            }
+        segments = network.get(mpnet.SEGMENTS)
+        if segments:
+            for seg in segments:
+                yield seg
+
+    def _validate_network_type(self, context, network_id):
+        our_types = [m_const.TYPE_MIDONET, m_const.TYPE_UPLINK]
+        network = self._core_plugin.get_network(context, network_id)
+        for seg in self._segments(network):
+            if seg[pnet.NETWORK_TYPE] in our_types:
+                return
+        LOG.warning(_LW("Incompatible network %s"), network)
+        raise n_exc.BadRequest(resource='router', msg='Incompatible network')
+
+    def _validate_router_gw_network(self, context, r):
+        ext_gw_info = r.get(l3.EXTERNAL_GW_INFO)
+        if ext_gw_info:
+            self._validate_network_type(context, ext_gw_info['network_id'])
+
     @log_helpers.log_method_call
     def create_router(self, context, router):
         with context.session.begin(subtransactions=True):
             r = super(MidonetL3ServicePlugin, self).create_router(context,
                                                                   router)
+            self._validate_router_gw_network(context, r)
             self.client.create_router_precommit(context, r)
 
         try:
@@ -87,13 +118,32 @@ class MidonetL3ServicePlugin(common_db_mixin.CommonDbMixin,
         with context.session.begin(subtransactions=True):
             r = super(MidonetL3ServicePlugin, self).update_router(context, id,
                                                                   router)
+            self._validate_router_gw_network(context, r)
             self.client.update_router_precommit(context, id, r)
 
-        self.client.update_router_postcommit(id, r)
+        try:
+            self.client.update_router_postcommit(id, r)
+            if r['status'] != m_const.ROUTER_STATUS_ACTIVE:
+                data = {'router': {'status': m_const.ROUTER_STATUS_ACTIVE}}
+                r = super(MidonetL3ServicePlugin,
+                        self).update_router(context, id, data)
+        except Exception as ex:
+            with excutils.save_and_reraise_exception():
+                LOG.error(_LE("Failed to update a router %(r_id)s in MidoNet: "
+                              "%(err)s"), {"r_id": id, "err": ex})
+                try:
+                    data = {'router': {'status': m_const.ROUTER_STATUS_ERROR}}
+                    super(MidonetL3ServicePlugin,
+                        self).update_router(context, id, data)
+                except Exception:
+                    LOG.exception(_LE("Failed to update a router "
+                                      "status %s"), id)
         return r
 
     @log_helpers.log_method_call
     def delete_router(self, context, id):
+        self._check_router_not_in_use(context, id)
+
         with context.session.begin(subtransactions=True):
             super(MidonetL3ServicePlugin, self).delete_router(context, id)
             self.client.delete_router_precommit(context, id)
@@ -102,9 +152,11 @@ class MidonetL3ServicePlugin(common_db_mixin.CommonDbMixin,
 
     @log_helpers.log_method_call
     def add_router_interface(self, context, router_id, interface_info):
+        by_port = bool(interface_info.get('port_id'))
         with context.session.begin(subtransactions=True):
             info = super(MidonetL3ServicePlugin, self).add_router_interface(
                 context, router_id, interface_info)
+            self._validate_network_type(context, info['network_id'])
             self.client.add_router_interface_precommit(context, router_id,
                                                        info)
 
@@ -116,7 +168,8 @@ class MidonetL3ServicePlugin(common_db_mixin.CommonDbMixin,
                           "error=%(err)r"),
                       {"info": info, "router_id": router_id, "err": ex})
             with excutils.save_and_reraise_exception():
-                self.remove_router_interface(context, router_id, info)
+                if not by_port:
+                    self.remove_router_interface(context, router_id, info)
 
         return info
 
@@ -164,6 +217,21 @@ class MidonetL3ServicePlugin(common_db_mixin.CommonDbMixin,
         with context.session.begin(subtransactions=True):
             fip = super(MidonetL3ServicePlugin, self).update_floatingip(
                 context, id, floatingip)
+            port_id = fip['port_id']
+            if port_id is not None:
+                port = self._core_plugin.get_port(context, port_id)
+                owner = port['device_owner']
+                # REVISIT(yamamoto): Empty owner is allowed for tempest and
+                # unit tests.
+                # NOTE(yamamoto): VIP is allowed for non MidoNet LB providers
+                if (owner and
+                   not owner.startswith(n_const.DEVICE_OWNER_COMPUTE_PREFIX)
+                   and owner != n_const.DEVICE_OWNER_LOADBALANCER):
+                    raise n_exc.UnsupportedPortDeviceOwner(
+                        op='floatingip association',
+                        port_id=port_id,
+                        device_owner=owner)
+
             self.client.update_floatingip_precommit(context, id, fip)
 
             # Update status based on association
@@ -173,5 +241,17 @@ class MidonetL3ServicePlugin(common_db_mixin.CommonDbMixin,
                 fip['status'] = n_const.FLOATINGIP_STATUS_ACTIVE
             self.update_floatingip_status(context, id, fip['status'])
 
-        self.client.update_floatingip_postcommit(id, fip)
+        try:
+            self.client.update_floatingip_postcommit(id, fip)
+        except Exception as ex:
+            with excutils.save_and_reraise_exception():
+                LOG.error(_LE("Failed to update a floating ip "
+                              "%(fip_id)s in MidoNet: "
+                              "%(err)s"), {"fip_id": id, "err": ex})
+                try:
+                    self.update_floatingip_status(
+                            context, id, n_const.FLOATINGIP_STATUS_ERROR)
+                except Exception:
+                    LOG.exception(_LE("Failed to update floating ip "
+                                      "status %s"), id)
         return fip
